# -*- coding: utf-8 -*-
"""
Helper module for using smtplib for secured connections

Author: Michael Ströder <michael@stroeder.com>

mainly a wrapper around smtplib.SMTP etc.
"""

__version__ = '0.2.0'

#-----------------------------------------------------------------------
# Import modules
#-----------------------------------------------------------------------

import socket
import smtplib
import email
import email.mime.text
import ssl
import hashlib
import urlparse

from sys import stderr

urlparse.uses_query.extend(['smtp', 'smtps'])

smtp_socket = socket.socket

#-----------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------

class SMTPTlsException(smtplib.SMTPException):
    """
    TLS-related exception raised during SMTP session
    """
    pass


class SMTPHostnameCheckFailed(SMTPTlsException):
    """
    RFC 6125 hostname check failed during SMTPS or STARTTLS
    """
    pass


class SMTPCertPinningCheckFailed(SMTPTlsException):
    """
    Cert-pinning check failed during SMTPS or STARTTLS
    """
    pass


class SMTP(smtplib.SMTP):
    """
    This class inherits from `smtplib.SMTP'.
    It mainly passes more arguments from its method calls
    `__init__()' and `starttls()' to function `ssl.wrap_socket()'.
    """
    default_port = smtplib.SMTP_PORT

    def __init__(
            self,
            host='',
            remote_addr=None,
            port=0,
            local_hostname=None,
            timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
            debug_level=0
        ):
        """
        `host' is the name against which TLS hostname check is
        performed. `remote_addr', if specified, is the
        hostname or IP address actually used by smtlib.SMTP to connect.
        All other arguments the same like for `smtplib.SMTP()'.
        """
        self.set_debuglevel(debug_level)
        smtplib.SMTP.__init__(
            self,
            host=remote_addr or host,
            port=port or self.default_port,
            local_hostname=local_hostname,
            timeout=timeout
        )
        # Save hostname to connect to for later TLS hostname check
        self._remote_host = host.strip().lower()

    def _get_socket(self, host, port, timeout, source_address=None):
        """
        """
        if self.debuglevel > 0:
            print>>stderr, 'connecting to %s:%s with %s.%s' % (
                host,
                port,
                smtp_socket.__module__,
                smtp_socket.__name__
            )
        err = None
        for res in socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM):
            af, socktype, proto, canonname, sa = res
            sock = None
            try:
                sock = smtp_socket(af, socktype, proto)
                if timeout is not socket._GLOBAL_DEFAULT_TIMEOUT:
                    sock.settimeout(timeout)
                if source_address:
                    sock.bind(source_address)
                sock.connect(sa)
                return sock

            except socket.error as _:
                err = _
                if sock is not None:
                    sock.close()
        if err is not None:
            raise err
        else:
            raise socket.error("getaddrinfo returns an empty list")

    def starttls(
            self,
            keyfile=None,
            certfile=None,
            cert_reqs=ssl.CERT_NONE,
            ssl_version=ssl.PROTOCOL_TLSv1,
            ca_certs=None,
            ciphers=None,
            expected_fingerprints=None,
        ):
        """
        send STARTTLS after set TLS options
        """
        self.ehlo_or_helo_if_needed()
        if not self.has_extn("starttls"):
            raise smtplib.SMTPException("STARTTLS extension not supported by server.")
        (resp, reply) = self.docmd("STARTTLS")
        if resp == 220:
            if not smtplib._have_ssl:
                raise RuntimeError("No SSL support included in this Python")
            tls_options = {
                'keyfile':keyfile,
                'certfile':certfile,
                'server_side':False,
                'cert_reqs':cert_reqs,
                'ssl_version':ssl_version,
                'ca_certs':ca_certs,
            }
            # key-word argument ciphers not support in Python 2.6 and older
            if ciphers:
                tls_options['ciphers'] = ciphers
            self.sock = ssl.wrap_socket(
                self.sock,
                **tls_options
            )
            self.file = smtplib.SSLFakeFile(self.sock)
            #
            self._tls_hostname_check(self.sock)
            self._check_fingerprints(self.sock, expected_fingerprints)
            # RFC 3207:
            # The client MUST discard any knowledge obtained from
            # the server, such as the list of SMTP service extensions,
            # which was not obtained from the TLS negotiation itself.
            self.helo_resp = None
            self.ehlo_resp = None
            self.esmtp_features = {}
            self.does_esmtp = 0
            # Resend EHLO after STARTTLS
            self.ehlo_or_helo_if_needed()
        return (resp, reply)

    def _check_fingerprints(self, sock, expected_fingerprints):
        """
        Implements simple certificate pinning by checking against a
        dictionary if expected fingerprints (hex-digest)
        """
        if not expected_fingerprints:
            # Nothing to do because there aren't any expected fingerprints given
            return
        # Get the server cert as raw DER blob
        cert_der = sock.getpeercert(binary_form=True)
        if not cert_der:
            # No server cert available
            raise SMTPCertPinningCheckFailed(
                'Could not get peer certificate to check fingerprint(s)!'
            )
        for hash_algo in expected_fingerprints.keys():
            h = hashlib.new(hash_algo)
            h.update(cert_der)
            h_hexdigest = h.hexdigest()
            if expected_fingerprints[hash_algo].replace(':', '') != h_hexdigest:
                raise SMTPCertPinningCheckFailed(
                    'Wrong fingerprint of peer certificate! Expected %s but got %s' % (
                        expected_fingerprints[hash_algo],
                        h_hexdigest
                    )
                )
        return # _check_fingerprints()

    def _tls_hostname_check(self, sock):
        peercert_dict = sock.getpeercert(binary_form=False)
        peercert_wildcard_list = []
        try:
            subject_dn_list = peercert_dict['subject']
        except KeyError:
            peercert_hostname_list = []
        else:
            peercert_hostname_list = []
            for rdn_list in subject_dn_list:
                for at, av in rdn_list:
                    if at == 'commonName':
                        av = av.strip().lower()
                        if av.startswith('*.'):
                            peercert_wildcard_list.append(av.encode('ascii'))
                        else:
                            peercert_hostname_list.append(av.encode('ascii'))
        peercert_ipaddr_list = []
        for altname_type, altname_data in peercert_dict.get('subjectAltName', []):
            if altname_type == 'DNS':
                peercert_hostname_list.append(
                    altname_data.strip().lower().encode('ascii')
                )
            if altname_type == 'IP': # FIX ME! Check if really 'IP' is the type!!
                peercert_ipaddr_list.append(altname_data.encode('ascii'))
        if not (
                self._remote_host in peercert_hostname_list or
                self._remote_host in peercert_ipaddr_list
            ):
            for w in peercert_wildcard_list:
                if self._remote_host.endswith(w[2:]):
                    break
            else:
                raise SMTPHostnameCheckFailed(
                    '%r not found in %r' % (self._remote_host, peercert_dict)
                )
        return # _tls_hostname_check()

    def send_simple_message(
            self,
            from_addr,
            to_addr,
            charset,
            headers,
            msg
        ):
        """
        Send a simple text/plain message
        """
        e_mail = email.mime.text.MIMEText(msg, 'plain', charset)
        e_mail.set_charset(charset)
        for header_name, header_value in headers:
            e_mail[header_name] = header_value
        self.sendmail(from_addr, to_addr, e_mail.as_string())


class SMTPS(SMTP):
    """
    SMTP over existing TLS connection aka SMTPS
    """
    default_port = smtplib.SMTP_SSL_PORT

    def __init__(
            self,
            host='',
            remote_addr=None,
            port=0,
            local_hostname=None,
            keyfile=None,
            certfile=None,
            cert_reqs=ssl.CERT_NONE,
            ssl_version=ssl.PROTOCOL_TLSv1,
            ca_certs=None,
            ciphers=None,
            expected_fingerprints=None,
            timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
            debug_level=0,
        ):
        self.keyfile = keyfile
        self.certfile = certfile
        self.cert_reqs = cert_reqs
        self.ssl_version = ssl_version
        self.ca_certs = ca_certs
        self.ciphers = ciphers
        self.expected_fingerprints = expected_fingerprints
        self._remote_host = host.strip().lower()
        SMTP.__init__(
            self,
            host=remote_addr or host,
            port=port or self.default_port,
            local_hostname=local_hostname,
            timeout=timeout,
            debug_level=debug_level,
        )

    def _get_socket(self, host, port, timeout):
        new_socket = SMTP._get_socket(self, host, port, timeout)
        tls_options = {
            'keyfile':self.keyfile,
            'certfile':self.certfile,
            'server_side':False,
            'cert_reqs':self.cert_reqs,
            'ssl_version':self.ssl_version,
            'ca_certs':self.ca_certs,
        }
        # key-word argument ciphers not support in Python 2.6 and older
        if self.ciphers:
            tls_options['ciphers'] = ciphers
        new_socket = ssl.wrap_socket(
            new_socket,
            **tls_options
        )
        self.file = smtplib.SSLFakeFile(new_socket)
        self._tls_hostname_check(new_socket)
        self._check_fingerprints(new_socket, self.expected_fingerprints)
        return new_socket


def smtp_connection(
        smtp_url,
        tls_args=None,
        local_hostname=None,
        timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
        debug_level=0
    ):
    """
    This function opens a SMTP connection specified by `smtp_url'.

    It also invokes `SMTP.starttls()' or `SMTP.login()' if URL contains
    appropriate parameters.
    """
    smtp_url_obj = urlparse.urlparse(smtp_url)
    tls_args = tls_args or {}
    query_obj = urlparse.parse_qs(
        smtp_url_obj.query,
        keep_blank_values=True,
        strict_parsing=False
    )
    remote_addr = query_obj.get('VIA', [None])[0]
    if smtp_url_obj.scheme == 'smtp':
        # Start clear-text connection
        smtp_conn = SMTP(
            host=smtp_url_obj.hostname,
            remote_addr=remote_addr,
            port=smtp_url_obj.port,
            local_hostname=local_hostname,
            timeout=timeout,
            debug_level=debug_level,
        )
        # Check whether STARTTLS was provided in URL
        if 'STARTTLS' in query_obj:
            smtp_conn.starttls(**tls_args)
    elif smtp_url_obj.scheme == 'smtps':
        kwargs = {
            'host':smtp_url_obj.hostname,
            'remote_addr':remote_addr,
            'port':smtp_url_obj.port,
            'local_hostname':local_hostname,
            'timeout':timeout,
            'debug_level':debug_level,
        }
        kwargs.update(tls_args)
        smtp_conn = SMTPS(**kwargs)
    else:
        raise ValueError('Unsupported URL scheme %r' % (smtp_url_obj.scheme))
    # Login if username and password were provided in URL
    if smtp_url_obj.username and smtp_url_obj.password:
        smtp_conn.login(smtp_url_obj.username, smtp_url_obj.password)
    return smtp_conn # smtp_connection()
